#include <graph/server/gui/chart/chart.h>
#include <graph/server/gui/mainwindow.h>
#include <qabstractseries.h>
#include <qlineseries.h>
#include <qmutex.h>
#include <QIcon>
#include <iostream>
#include <limits>

#define MIN(a, b) (a < b ? a : b)
#define MAX(a, b) (a > b ? a : b)

namespace Graph::GUI {
ChartWidget::ChartWidget(QWidget* parent, const chart_spec_t& spec) : QListWidgetItem(), spec(spec) {
  view = new QChartView(parent);
  chart = new QChart();

  // chart
  if (spec.legend)
    chart->legend()->show();
  else
    chart->legend()->hide();

  chart->setTitle(spec.title);

  // axis
  x = new QValueAxis();
  y = new QValueAxis();

  if (!spec.axis.x.auto_limit)
    x->setRange(spec.axis.x.min, spec.axis.x.max);
  if (!spec.axis.y.auto_limit)
    y->setRange(spec.axis.y.min, spec.axis.y.max);

  x->setTitleText(spec.axis.x.label);
  y->setTitleText(spec.axis.y.label);

  chart->addAxis(x, Qt::AlignBottom);
  chart->addAxis(y, Qt::AlignLeft);

  view->setRenderHint(QPainter::Antialiasing);
  view->setChart(chart);
  view->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
  setSizeHint(view->sizeHint() * 4);

  // context menu
  view->setContextMenuPolicy(Qt::ContextMenuPolicy::CustomContextMenu);
  view->connect(view, &QChartView::customContextMenuRequested, view, [this](const QPoint& pos) {
    ContextMenu menu(nullptr, this);
    menu.exec(view->mapToGlobal(pos));
  });
}

ChartWidget::~ChartWidget() {
  // delete view;
  // delete chart;
  // delete x;
  // delete y;
}

void ChartWidget::addData(const char* seriesName, const point_t* points, int count) {
  QMutexLocker locker(&mutex);

  QString name(seriesName);

  double minX = spec.keep_all ? x->min() : std::numeric_limits<double>::max();
  double minY = spec.keep_all ? y->min() : std::numeric_limits<double>::max();
  double maxX = spec.keep_all ? x->max() : std::numeric_limits<double>::min();
  double maxY = spec.keep_all ? y->max() : std::numeric_limits<double>::min();

  for (QAbstractSeries* series : chart->series()) {
    if (series->name() == name) {
      if (!spec.keep_all) {
        if (count >= spec.keep_limit) {
          ((QLineSeries*)series)->clear();
        } else if (((QLineSeries*)series)->count() + count > spec.keep_limit) {
          ((QLineSeries*)series)->removePoints(0, ((QLineSeries*)series)->count() - (spec.keep_limit - count));
        }

        for (QPointF point : ((QLineSeries*)series)->points()) {
          if (spec.axis.x.auto_limit) {
            minX = MIN(minX, point.x());
            maxX = MAX(maxX, point.x());
          }
          if (spec.axis.y.auto_limit) {
            minY = MIN(minY, point.y());
            maxY = MAX(maxY, point.y());
          }
        }
      }

      for (int i = spec.keep_all ? 0 : MAX(0, count - spec.keep_limit); i < count; i++) {
        if (decimateState >= spec.decimate)
          decimateState = 0;
        if (decimateState++ % spec.decimate != 0)
          continue;

        if (spec.axis.x.auto_limit) {
          minX = MIN(minX, points[i].x);
          maxX = MAX(maxX, points[i].x);
        }
        if (spec.axis.y.auto_limit) {
          minY = MIN(minY, points[i].y);
          maxY = MAX(maxY, points[i].y);
        }

        *((QLineSeries*)series) << QPointF(points[i].x, points[i].y);
      }

      if (spec.axis.x.auto_limit)
        x->setRange(minX, maxX);
      if (spec.axis.y.auto_limit)
        y->setRange(minY, maxY);

      break;
    }
  }

  view->update();
}

void ChartWidget::clear() {
  QMutexLocker locker(&mutex);

  for (QAbstractSeries* series : chart->series()) {
    ((QLineSeries*)series)->clear();
  }
  x->setRange(spec.axis.x.min, spec.axis.x.max);
  y->setRange(spec.axis.y.min, spec.axis.y.max);
  decimateState = 0;
  view->update();
}

ContextMenu::ContextMenu(QWidget* parent, ChartWidget* chartWidget) : QMenu(parent) {
  remove = new QAction(tr("&Remove"), this);
  connect(remove, &QAction::triggered, this, [chartWidget]() { MainWindow::getInstance()->removeChart(chartWidget); });
  addAction(remove);

  clear = new QAction(tr("&Clear"), this);
  connect(clear, &QAction::triggered, this, [chartWidget]() { chartWidget->clear(); });
  addAction(clear);

  edit = new QAction(tr("&Edit"), this);
  connect(edit, &QAction::triggered, this, []() {
    // TODO: implement chart editing
  });
  addAction(edit);
}

ContextMenu::~ContextMenu() {}
} // namespace Graph::GUI